import contextlib
import ldap
from ldap.ldapobject import ReconnectLDAPObj
from authserv import model


class Error(Exception):
    pass


class UserDb(model.UserDb):

    ldap_attrs = [
        'userPassword',
        'totpSecret',
        'appSpecificPassword',
        ]

    def __init__(self, service_map, ldap_uri, ldap_bind_dn, ldap_bind_pw):
        self.service_map = service_map
        self.ldap_uri = ldap_uri
        self.ldap_bind_dn = ldap_bind_dn
        self.ldap_bind_pw = ldap_bind_pw

    @contextlib.contextmanager
    def _conn(self):
        c = ReconnectLDAPObj(self.ldap_uri)
        c.protocol_version = ldap.VERSION3
        c.simple_bind_s(self.ldap_bind_dn, self.ldap_bind_pw)
        yield c
        c.unbind_s()

    def _query_user(self, username, service):
        ldap_params = self.service_map.get(service)
        if not ldap_params:
            return None

        if callable(ldap_params['dn']):
            basedn = ldap_params['dn'](username)
        else:
            basedn = ldap_params['dn']

        with self._conn() as c:
            result = c.search_s(
                basedn,
                ldap_params.get('scope', ldap.SCOPE_SUBTREE),
                ldap_params['filter'].replace('%s', username),
                self.ldap_attrs)

        if not result:
            return None
        if len(result) > 1:
            raise Error('too many results from LDAP')

        return User(username, result[0][0], result[0][1])

    def get_user(self, username, service):
        try:
            return User(username)
        except (Error, ldap.LDAPError):
            return None


class User(model.User):

    def __init__(self, username, dn, data):
        self._username = username
        self._dn = dn
        self._otp_enabled = False
        self._asps = []
        for key, values in data.iteritems():
            if key == 'userPassword':
                self._password = values[0]
                if self._password.startswith('{crypt}'):
                    self._password = self._password[7:]
            elif key == 'totpSecret':
                self._otp_enabled = True
                self._totp_secret = values[0]
            elif key == 'appSpecificPassword':
                self.asps = [v.split(':', 2) for v in values]

    def otp_enabled(self):
        return self._otp_enabled

    app_specific_passwords_enabled = otp_enabled

    def get_name(self):
        return self._username

    def get_totp_key(self):
        return self._totp_secret

    def get_app_specific_passwords(self, service):
        return [x[2] for x in self._asps if x[0] == service]

    def get_password(self):
        return self._password
